import fakeredis
from unittest.mock import Mock
from atomicpuppy.atomicpuppy import RedisCounter
from .fakehttp import SpyLog
from redis import TimeoutError

class When_redis_contains_no_information_for_a_stream:

    def given_an_empty_redis_counter(self):
        self.counter = RedisCounter(fakeredis.FakeStrictRedis(), instance="fred")

    def it_should_return_negative_one(self):
        assert(self.counter["my-stream"] == -1)


class When_redis_contains_a_last_read_value_for_a_stream:

    last_read = 92837

    def given_a_redis_instance_with_a_last_read_event(self):
        self.redis = fakeredis.FakeStrictRedis()
        self.redis.set('urn:atomicpuppy:fred:my-stream:position', self.last_read)

    def it_should_return_the_last_read_value(self):
        ctr = RedisCounter(self.redis, instance="fred")
        assert(ctr["my-stream"] == self.last_read)

    def cleanup_redis(self):
        self.redis.flushall()


class When_setting_the_last_read_value:

    def given_an_empty_redis_instance(self):
        self.redis = fakeredis.FakeStrictRedis()
        self.ctr = RedisCounter(self.redis, "instance")

    def because_we_record_an_event(self):
        self.ctr["my-stream"] = 10

    def it_should_be_persisted_in_redis(self):
        assert(self.ctr["my-stream"] == 10)

    def cleanup_redis(self):
        self.redis.flushall()



import aiohttp
import asyncio
import datetime
from enum import Enum
from .errors import *
import json
import logging
from collections import namedtuple, defaultdict
import platform
import pybreaker
from retrying import retry
import random
import redis
import requests
from uuid import UUID
import yaml
from concurrent.futures import TimeoutError

SubscriptionConfig = namedtuple('SubscriptionConfig', ['streams',
                                                       'counter_factory',
                                                       'instance_name',
                                                       'host',
                                                       'port',
                                                       'timeout'])


class Event:

    def __init__(self, id, type, data, stream, sequence):
        self.id = id
        self.type = type
        self.data = data
        self.stream = stream
        self.sequence = sequence

    def __str__(self):
        return "{}-{} ({}): {}".format(self.type, self.sequence, self.id, self.data)


class EventCounterCircuitBreaker(pybreaker.CircuitBreakerListener):

    def __init__(self):
        self._log = logging.getLogger(__name__)

    def before_call(self, cb, func, *args, **kwargs):
        "Called before the circuit breaker `cb` calls `func`."
        pass

    def state_change(self, cb, old_state, new_state):
        if(new_state.name == 'open'):
            self._log.error("Opening event counter circuit")
        elif(new_state.name == 'closed'):
            self._log.info("Event counter circuit closed")
        pass

    def failure(self, cb, exc):
        "Called when a function invocation raises a system error."
        pass

    def success(self, cb):
        "Called when a function invocation succeeds."
        pass


redis_circuit = pybreaker.CircuitBreaker(
    fail_max=20, reset_timeout=60, listeners=[
        EventCounterCircuitBreaker()
    ]
)


class SubscriberInfo:

    def __init__(
            self,
            stream,
            uri,
            last_read=-1):
        self._stream = stream
        self._uri = uri
        self._last_read = last_read

    @property
    def stream(self):
        return self._stream

    @property
    def last_read(self):
        return self._last_read

    @property
    def uri(self):
        return self._uri


class StreamReader:

    def __init__(self, queue, stream_name, loop, instance_name, subscriptions_store, timeout, nosleep=False):
        self._fetcher = StreamFetcher(None, loop=loop, nosleep=nosleep, timeout=timeout)
        self._queue = queue
        self._loop = loop
        self._stream = stream_name
        self._subscriptions = subscriptions_store
        self.logger = logging.getLogger('atomicpuppy.stream-reader@{}/{}'.format(instance_name, stream_name))

    @asyncio.coroutine
    def start_consuming(self, once=False):
        self._running = True
        while(self._running):
            try:
                yield from self.consume()
                if(once or not self._running):
                    break
                yield from asyncio.sleep(5)
            except UrlError as e:
                self.logger.exception(
                    "Failed to make a request with uri %s",
                    e.uri)
                self.stop()
            except HttpClientError as e:
                self.logger.exception(
                    "Received bad http response with status %d from %s",
                    e.status,
                    e.uri)
                self.stop()
            except asyncio.CancelledError:
                pass

    def stop(self):
        self._running = False
        self.logger.info("Stopping")
        self._fetcher.stop()

    @asyncio.coroutine
    def consume(self):
        subscription = self._subscriptions.get(self._stream)
        if(subscription.last_read > -1):
            self.logger.debug(
                "Last read event is %d - seeking last read on page %s",
                subscription.last_read,
                subscription.uri)
            yield from self.seek_on_page(subscription.uri)
        else:
            r = yield from self._fetcher.fetch(subscription.uri)
            js = yield from r.json()

            last = self._get_link(js, "last")
            if(last):
                self.logger.debug(
                    "No last read event, skipping to last page %s",
                    last)
                yield from self._walk_from_last(last)
            else:
                self.logger.debug(
                    "No last read event, yielding events from this page %s",
                    subscription.uri)
                yield from self._raise_page_events(js)

    @asyncio.coroutine
    def seek_on_page(self, uri):
        self.logger.debug("Looking for last read event on page %s", uri)
        r = yield from self._fetcher.fetch(uri)
        js = yield from r.json()
        yield from self.seek_to_last_read(js)

    @asyncio.coroutine
    def _walk_from_last(self, prev_uri):
        uri = prev_uri
        while True:
            self.logger.debug("walking backwards from %s", uri)
            r = yield from self._fetcher.fetch(uri)
            js = yield from r.json()
            if js["entries"]:
                self.logger.debug("raising events from page %s", uri)
                yield from self._raise_page_events(js)
            prev_uri = self._get_link(js, "previous")
            if not prev_uri:
                # loop until there's a prev link, otherwise it means
                # we are at the end or we are fetching an empty page
                self.logger.debug(
                    "back-walk completed, new polling uri is %s",
                    uri)
                self._subscriptions.update_uri(self._stream, uri)
                break

            self.logger.debug("Continuing to previous page %s", prev_uri)
            uri = prev_uri

    @asyncio.coroutine
    def _raise_page_events(self, js):
        subscription = self._subscriptions.get(self._stream)
        stack = []
        for e in js['entries']:
            if(e["positionEventNumber"] > subscription.last_read):
                stack.append(e)
        while(stack):
            evt = self._make_event(stack.pop())
            if(evt):
                yield from self._queue.put(evt)
                self._subscriptions.update_sequence(self._stream, evt.sequence)

    @asyncio.coroutine
    def seek_to_last_read(self, js):
        subscription = self._subscriptions.get(self._stream)
        for e in js["entries"]:
            if(e["positionEventNumber"] <= subscription.last_read):
                self.logger.debug(
                    "Found last read event on current page, raising events")
                yield from self._raise_page_events(js)
                prev = self._get_link(js, "previous")
                if(prev):
                    yield from self._walk_from_last(prev)
                    return

        nxt = self._get_link(js, "next")
        if(nxt):
            yield from self.seek_on_page(nxt)

    def _get_link(self, js, rel):
        for link in js["links"]:
            if(link["relation"] == rel):
                return link["uri"]

    def _make_event(self, e):
        try:
            data = json.loads(e["data"], encoding='UTF-8')
        except KeyError:
            # Eventstore allows events with no `data` to be posted. If that
            # happens, then atomicpuppy doesn't know what to do
            self.logger.warning("No `data` key found on event {}".format(e))
            return None
        except ValueError:
            self.logger.error("Failed to parse json data for %s message %s",
                              e.get("eventType"), e.get("eventId"))
            return None
        type = e["eventType"]
        id = UUID(e["eventId"])
        stream = e["positionStreamId"]
        sequence = e["positionEventNumber"]
        return Event(id, type, data, stream, sequence)


class state(Enum):
    green = 0
    amber = 1
    red = 2


class StreamFetcher:

    def __init__(self, policy, loop, timeout, nosleep=False):
        self._policy = policy
        self._loop = loop
        self._log = logging.getLogger(__name__)
        self._state = state.green
        self._running = True
        self._sleep = None
        self._nosleep = nosleep
        self._exns = set()
        self._timeout = timeout

    def stop(self):
        self._running = False
        if(self._sleep):
            self._sleep.cancel()

    def sleeps(self, uri):
        max_time = 60
        attempts = 0
        last_time = 0
        self._state = state.amber
        while(True):
            if(last_time < max_time):
                attempts = attempts + 1
            elif self._state == state.amber:
                self._state = state.red
                self._log.error(
                    "Stream fetcher has failed to connect to eventstore at %s in the last %d attempts",
                    uri,
                    attempts)
            last_time = (2**attempts)
            jitter = random.uniform(-0.5, 0.5)
            yield (last_time) + jitter

    def log(self, e, uri):
        if(type(e) not in self._exns):
            self._log.warn(
                "Error occurred while requesting %s",
                uri,
                exc_info=True)
            self._exns.add(type(e))

    @asyncio.coroutine
    def sleep(self, delay):
        if(self._nosleep):
            return
        self._sleep = asyncio.futures.Future(loop=self._loop)
        self._sleep._loop.call_later(
            delay,
            self._sleep._set_result_unless_cancelled,
            None)
        self._log.debug("retrying fetch in %d seconds", delay)
        yield from self._sleep

    @asyncio.coroutine
    def fetch(self, uri):
        sleep_times = self.sleeps(uri)
        self._exns = set()
        self._state = state.green
        for s in sleep_times:
            headers = {"Accept": "application/json"}
            params = {"embed": "body"}
            try:
                response = yield from asyncio.wait_for(
                    aiohttp.request(
                        'GET', uri,
                        params=params,
                        headers=headers,
                        loop=self._loop,
                        connector=aiohttp.TCPConnector(
                            resolve=True,
                            loop=self._loop,
                        )
                    ),
                    self._timeout,
                    loop=self._loop
                )
                if(response.status == 200):
                    return response
                if response.status in (404, 408):
                    raise HttpServerError(uri, response.status)
                if(response.status >= 400 and response.status <= 499):
                    raise HttpClientError(uri, response.status)
                if(response.status >= 500 and response.status <= 599):
                    raise HttpServerError(uri, response.status)
            except ValueError as e:
                raise UrlError(e)
            except (aiohttp.errors.ClientError, aiohttp.errors.DisconnectedError, aiohttp.errors.ClientResponseError) as e:
                self.log(e, uri)
                yield from self.sleep(s)
            except HttpServerError as e:
                self.log(e, uri)
                yield from self.sleep(s)
            except TimeoutError as e:
                self.log(e, uri)
                yield from self.sleep(s)



class EventRaiser:

    def __init__(self, queue, counter, callback, loop=None):
        self._queue = queue
        self._callback = callback
        self._counter = counter
        self._loop = loop or asyncio.get_event_loop()
        self._logger = logging.getLogger(__name__)

    def stop(self):
        self._is_running = False
        self._logger.info("Stopping")

    @asyncio.coroutine
    def start(self):
        self._is_running = True
        while(self._loop.is_running() and self._is_running):
            try:
                msg = yield from asyncio.wait_for(self._queue.get(),
                                                  timeout=1,
                                                  loop=self._loop)
                if not msg:
                    continue
                self._callback(msg)
                try:
                    self._counter[msg.stream] = msg.sequence
                except CircuitBreakerError:
                    pass
                except RedisError:
                    self._logger.warn("Failed to persist last read event")
            except RejectedMessageException:
                self._logger.warn("%s message %s was rejected and has not been processed",
                                  msg.type, msg.id)
            except(TimeoutError):
                pass
            except:
                self._logger.exception("Failed to process message %s", msg)


    @asyncio.coroutine
    def consume_events(self):
        self._is_running = True
        while(self._loop.is_running() and self._is_running):
            try:
                msg = yield from asyncio.wait_for(self._queue.get(),
                                                  timeout=1,
                                                  loop=self._loop)
                if not msg:
                    self._is_running = False
                    return
                self._callback(msg)
                try:
                    self._counter[msg.stream] = msg.sequence
                except CircuitBreakerError:
                    pass
                except RedisError:
                    self._logger.warn("Failed to persist last read event")
            except RejectedMessageException:
                self._logger.warn("%s message %s was rejected and has not been processed",
                                  msg.type, msg.id)
            except(TimeoutError):
                self._is_running = False
            except:
                self._logger.exception("Failed to process message %s", msg)
                self._is_running = False


class EventCounter:

    def __setitem__(self, stream, event):
        raise NotImplementedError("setitem")

    def __getitem__(self, stream):
        raise NotImplementedError("getitem")


class RedisCounter(EventCounter):

    def __init__(self, redis, instance):
        self._redis = redis
        self._instance_name = instance
        self._logger = logging.getLogger(__name__)

    @retry(wait_exponential_multiplier=1000, wait_exponential_max=10000, stop_max_delay=60000)
    def __getitem__(self, stream):
        self._logger.debug("Fetching last read event for stream "+stream)
        key = self._key(stream)
        val = self._redis.get(key)
        if(val):
            val = int(val)
            self._logger.info(
                "Last read event for stream %s is %d",
                stream,
                val)
            return val
        return -1

    @redis_circuit
    def __setitem__(self, stream, val):
        key = self._key(stream)
        self._redis.set(key, val)

    def _key(self, stream):
        return "urn:atomicpuppy:"+self._instance_name+":"+stream+":position"


class StreamConfigReader:

    def __init__(self):
        pass

    def read(self, config_file):
        cfg = None
        if isinstance(config_file, dict):
            cfg = config_file.get('atomicpuppy')
        elif isinstance(config_file, str):
            with open(config_file) as file:
                cfg = yaml.load(file).get('atomicpuppy')
        else:
            cfg = yaml.load(config_file).get('atomicpuppy')


        streams = []
        instance = cfg.get('instance') or platform.node()
        for stream in cfg.get("streams"):
            streams.append(stream)
        ctr = self._make_counter(cfg, instance)
        return SubscriptionConfig(streams=streams,
                                  counter_factory=ctr,
                                  instance_name=instance,
                                  host=cfg.get("host") or 'localhost',
                                  port=cfg.get("port") or 2113,
                                  timeout=cfg.get("timeout") or 20)

    def _make_counter(self, cfg, instance):
        ctr = cfg.get('counter')
        if(not ctr):
            return (lambda: defaultdict(lambda: -1))
        if(ctr["redis"]):
            return lambda: RedisCounter(
                redis.StrictRedis(port=ctr["redis"].get("port"),
                                  host=ctr["redis"].get("host")),
                instance)


class SubscriptionInfoStore:

    """
    This class acts like an in-memory buffer of subscriptions.
    It keeps track of subscriptions based on the stream_name
    with dynamic parts replaced.

    It's needed so that new SubscriberInfo objects can be created
    before a consumption when dynamic streams produce a new stream
    name.
    """

    def __init__(self, config, counter):
        self.subscriptions = {}
        self.config = config
        self.counter = counter
        self._logger = logging.getLogger("SubscriptionInfoStore")

    def get(self, stream_name):
        parsed_name = self._parse(stream_name)
        if parsed_name not in self.subscriptions.keys():
            last_read_for_stream = self.counter[parsed_name]
            self._logger.info("Creating new SubscriberInfo for {} from event number {}".format(parsed_name, last_read_for_stream))
            s = SubscriberInfo(parsed_name, uri=self._build_uri(parsed_name, last_read_for_stream), last_read=last_read_for_stream)
            self.subscriptions[parsed_name] = s

        return self.subscriptions[parsed_name]

    def update_sequence(self, stream_name, sequence):
        # the input stream_name can be the unparsed name
        s = self.get(stream_name)
        new_subscription = SubscriberInfo(
            stream=s._stream,
            uri=s.uri,
            last_read=sequence)
        old_subscription = self.subscriptions[s._stream]
        assert(old_subscription.last_read < sequence)
        self.subscriptions[s._stream] = new_subscription

    def update_uri(self, stream_name, uri):
        # the input stream_name can be the unparsed name
        s = self.get(stream_name)
        new_subscription = SubscriberInfo(
            stream=s._stream,
            uri=uri,
            last_read=s.last_read)
        self.subscriptions[s._stream] = new_subscription

    def _build_uri(self, stream_name, last_read_for_stream=-1):
        if last_read_for_stream < 0:
            last_read_for_stream = 0

        return 'http://{}:{}/streams/{}/{}/forward/20'.format(
            self.config.host, self.config.port, stream_name, last_read_for_stream)

    def _parse(self, stream_name):
        if '#date#' not in stream_name:
            return stream_name

        today = datetime.date.today().isoformat()
        return stream_name.replace("#date#", today)


class EventStoreJsonEncoder(json.JSONEncoder):

    def default(self, o):
        try:
            if(isinstance(o, UUID)):
                return str(o)
            return json.JSONEncoder.default(self, o)
        except e:
            logging.error(type(o))


class EventPublisher:

    def __init__(self, host, port):
        self._es_url = '{}:{}'.format(host, port)
        self._logger = logging.getLogger("EventPublisher")

    @retry(wait_exponential_multiplier=1000, wait_exponential_max=10000, stop_max_delay=30000,
           retry_on_result=lambda r: r.status_code >= 500)
    def post(self, event, correlation_id=None):
        uri = 'http://{}/streams/{}'.format(self._es_url, event.stream)
        headers = {'ES-EventType': str(event.type), 'ES-EventId': str(event.id), 'Content-Type': 'application/json'}
        extra = dict(event.__dict__)
        extra.update(headers)
        extra.update({
            'es_uri': uri,
            'method': 'POST'
        })
        self._logger.info("Posting event {} to {}. Headers: {}".format(event.__dict__, uri, headers), extra=extra)
        if(correlation_id):
            event.data["correlation_id"] = correlation_id
        r = requests.post(
            uri,
            headers=headers,
            data=EventStoreJsonEncoder().encode(event.data),
            timeout=0.5)
        if r.status_code:
            extra.update({
                'status_code': r.status_code
            })
            self._logger.info("Received status code {status_code} from EventStore POST.".format(**extra), extra=extra)
        return r



import io
from atomicpuppy.atomicpuppy import *
import platform


class When_reading_a_config_file:

    def given_a_config_file(self):
        self._file = io.BytesIO(b"""
                atomicpuppy:
                    host: eventstore.local
                    instance: my_stream_reader
                    port: 999
                    streams:
                        - foo
                        - bar
                        - baz
                        - dynamic_#date#
                    counter:
                        redis:
                            host: localhost
                            port: 1234
                """)

    def because_we_read_the_file(self):
        self.reader = StreamConfigReader()
        with self._file as f:
            self.result = self.reader.read(f)

    def it_should_contain_four_streams(self):
        assert(len(self.result.streams) == 4)

    def it_should_have_the_correct_host(self):
        assert(self.result.host == "eventstore.local")

    def it_should_have_the_correct_port(self):
        assert(self.result.port == 999)

    def it_should_have_the_correct_stream_id(self):
        assert(self.result.streams[0] == "foo")
        assert(self.result.streams[1] == "bar")
        assert(self.result.streams[2] == "baz")
        assert(self.result.streams[3] == "dynamic_#date#")

    def it_should_have_the_correct_instance_name(self):
        assert(self.result.instance_name == "my_stream_reader")


class When_the_host_is_not_specified:

    def given_a_config_file_with_no_host(self):
        self._file = io.BytesIO(b"""
                atomicpuppy:
                    port: 1234
                    streams:
                        - foo
                        - bar
                        - baz
                """)

    def because_we_read_the_file(self):
        self.reader = StreamConfigReader()
        with self._file as f:
            self.result = self.reader.read(f)

    def it_should_contain_three_streams(self):
        assert(len(self.result.streams) == 3)

    def it_should_have_the_correct_host(self):
        assert(self.result.host == "localhost")

    def it_should_have_the_correct_port(self):
        assert(self.result.port == 1234)


class When_the_port_is_not_specified:

    def given_a_config_file_with_no_host(self):
        self._file = io.BytesIO(b"""
                atomicpuppy:
                    streams:
                        - foo
                        - bar
                        - baz
                """)

    def because_we_read_the_file(self):
        self.reader = StreamConfigReader()
        with self._file as f:
            self.result = self.reader.read(f)

    def it_should_contain_three_streams(self):
        assert(len(self.result.streams) == 3)

    def it_should_have_the_correct_port(self):
        assert(self.result.port == 2113)


class When_the_instance_name_is_not_specified:

    def given_a_config_file_with_no_host(self):
        self._file = io.BytesIO(b"""
                atomicpuppy:
                    port: 1234
                    streams:
                        - foo
                        - bar
                        - baz
                """)

    def because_we_read_the_file(self):
        self.reader = StreamConfigReader()
        with self._file as f:
            self.result = self.reader.read(f)

    def it_should_default_to_the_hostname(self):
        assert(self.result.instance_name == platform.node())


class When_the_config_specifies_a_redis_counter:

    def given_a_config_file(self):
        self._file = io.BytesIO(b"""
                atomicpuppy:
                    host: eventstore.local
                    port: 999
                    streams:
                        - foo
                        - bar
                        - baz
                    counter:
                        redis:
                            host: localhost
                            port: 1234
                """)

    def because_we_read_the_file(self):
        self.reader = StreamConfigReader()
        with self._file as f:
            result = self.reader.read(f)
        self.ctr = result.counter_factory()

    def it_should_return_a_redis_counter(self):
        assert(isinstance(self.ctr, RedisCounter))



from .atomicpuppy import (
    StreamReader,
    StreamConfigReader,
    EventRaiser,
    Event,
    EventPublisher,
    EventStoreJsonEncoder,
    SubscriptionInfoStore
)
from .errors import *

import asyncio


class AtomicPuppy:

    running = False

    def __init__(self, cfg_file, callback, loop=None):
        self.config = StreamConfigReader().read(cfg_file)


        self.callback = callback
        self._loop = loop or asyncio.get_event_loop()
        self._queue = asyncio.Queue(maxsize=20, loop=self._loop)

    def start(self, run_once=False):
        c = self.counter = self.config.counter_factory()
        self._messageProcessor = EventRaiser(self._queue,
                                             c,
                                             self.callback,
                                             self._loop)
        subscription_info_store = SubscriptionInfoStore(self.config, c)
        self.readers = [
            StreamReader(
                queue=self._queue,
                stream_name=s,
                loop=self._loop,
                instance_name=self.config.instance_name,
                subscriptions_store=subscription_info_store,
                timeout=self.config.timeout)
            for s in self.config.streams
        ]
        self.tasks = [s.start_consuming(once=run_once) for s in self.readers]
        if run_once:
            self.tasks.append(self._messageProcessor.consume_events())
        else:
            self.tasks.append(self._messageProcessor.start())
        return asyncio.gather(*self.tasks, loop=self._loop)

    def stop(self):
        for s in self.readers:
            s.stop()
        self._messageProcessor.stop()



import asyncio
from atomicpuppy import EventRaiser, RejectedMessageException
from atomicpuppy.atomicpuppy import Event
from .fakehttp import SpyLog
from concurrent.futures import TimeoutError
from uuid import uuid4


class When_an_event_is_processed:

    the_message = None
    event_recorder = {}
    sequence_no = 43

    def given_an_event_raiser(self):
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        self.message_id = uuid4()

        self.queue = asyncio.Queue(loop=self._loop)
        self.message_processor = EventRaiser(self.queue,
                                                  self.event_recorder,
                                                  lambda e: self.process_message(e),
                                                  self._loop)

    def because_we_add_a_message(self):
        msg = Event(self.message_id, "type", {}, "stream", self.sequence_no)
        asyncio.async(self.send_message(msg), loop=self._loop)
        self._loop.run_until_complete(self.message_processor.start())

    def it_should_have_sent_the_message(self):
        assert(self.the_message.id == self.message_id)

    def it_should_have_recorded_the_event(self):
        assert(self.event_recorder["stream"] == self.sequence_no)

    @asyncio.coroutine
    def send_message(self, e):
        yield from self.queue.put(e)

    def process_message(self, e):
        self.the_message = e
        self.message_processor.stop()


class When_an_event_is_processed_by_running_once:

    the_message = None
    event_recorder = {}
    sequence_no = 43

    def given_an_event_raiser(self):
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        self.message_id = uuid4()

        self.queue = asyncio.Queue(loop=self._loop)
        self.message_processor = EventRaiser(self.queue,
                                                  self.event_recorder,
                                                  lambda e: self.process_message(e),
                                                  self._loop)

    def because_we_add_a_message(self):
        msg = Event(self.message_id, "type", {}, "stream", self.sequence_no)
        asyncio.async(self.send_message(msg), loop=self._loop)
        self._loop.run_until_complete(self.message_processor.consume_events())

    def it_should_have_sent_the_message(self):
        assert(self.the_message.id == self.message_id)

    def it_should_have_recorded_the_event(self):
        assert(self.event_recorder["stream"] == self.sequence_no)

    @asyncio.coroutine
    def send_message(self, e):
        yield from self.queue.put(e)

    def process_message(self, e):
        self.the_message = e



class When_a_message_is_rejected:

    event_recorder = {}

    def given_an_event_raiser(self):
        self._log = SpyLog()
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.message_id = uuid4()
        self.queue = asyncio.Queue(loop=self._loop)

        self.event_raiser = EventRaiser(
            self.queue,
            self.event_recorder,
            lambda e: self.process_message(e),
            self._loop
        )

    def because_we_process_a_message(self):
        with(self._log.capture()):
            msg = Event(self.message_id, "message-type", {}, "stream", 2)
            asyncio.async(self.send_message(msg), loop=self._loop)
            self._loop.run_until_complete(self.event_raiser.start())

    def it_should_log_a_warning(self):
        m = "message-type message "+str(self.message_id) \
            +" was rejected and has not been processed"
        assert(any(r.message == m for r in self._log.warnings))

    def process_message(self, e):
        self.event_raiser.stop()
        raise RejectedMessageException()

    @asyncio.coroutine
    def send_message(self, e):
        yield from self.queue.put(e)


class When_a_message_raises_an_unhandled_exception:

    event_recorder = {}

    def given_an_event_raiser(self):
        self._log = SpyLog()
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.message_id = uuid4()
        self.queue = asyncio.Queue(loop=self._loop)

        self.event_raiser = EventRaiser(
            self.queue,
            self.event_recorder,
            lambda e: self.process_message(e),
            self._loop
        )

    def because_we_process_a_message(self):
        with(self._log.capture()):
            msg = Event(self.message_id, "message-type", {}, "stream", 2)
            asyncio.async(self.send_message(msg), loop=self._loop)
            self._loop.run_until_complete(self.event_raiser.start())

    def it_should_log_an_error(self):
        m = "Failed to process message "
        assert(any(r.message.startswith(m) for r in self._log.errors))

    def process_message(self, e):
        self.event_raiser.stop()
        raise NotImplemented("This handler is not here")

    @asyncio.coroutine
    def send_message(self, e):
        yield from self.queue.put(e)





import aiohttp
import asyncio
import datetime
import logging
import os
from unittest.mock import patch
from uuid import UUID, uuid4

import fakeredis
from freezegun import freeze_time

from atomicpuppy.atomicpuppy import (
    StreamReader, SubscriptionInfoStore, SubscriptionConfig, RedisCounter
)
from .fakehttp import FakeHttp, SpyLog

SCRIPT_PATH = os.path.dirname(__file__)


class StreamReaderContext:

    _loop = None
    _events = None
    _host = 'eventstore.local'
    _port = 2113

    def __init__(self):
        self.counter = RedisCounter(fakeredis.FakeStrictRedis(), "test-instace-{}".format(uuid4()))

    def given_an_event_loop(self):
        logging.basicConfig(filename='example.log', level=logging.DEBUG)
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

        self.http = FakeHttp(self._loop)
        self._queue = asyncio.Queue(loop=self._loop)

    def subscribe_and_run(self, stream, last_read=-1, nosleep=False):
        self.subscribeTo(stream, last_read, nosleep)
        self.run_the_reader()

    def run_the_reader(self):
        mock = self.http.getMock()
        with patch("aiohttp.request", new=mock):
            self._loop.run_until_complete(
                self._reader.start_consuming(once=True)
            )

    def create_counter(self):
        return lambda: self.counter

    def subscribeTo(self, stream, last_read, nosleep=False):
        assert(last_read is not None)
        config = SubscriptionConfig(
            streams=None,
            counter_factory=self.create_counter,
            instance_name='foo',
            host=self._host,
            port=self._port,
            timeout=20)

        subscriptions_store = SubscriptionInfoStore(config, self.counter)
        if last_read != -1:
            self.counter[stream] = last_read
        self._reader = StreamReader(
            queue=self._queue,
            stream_name=stream,
            loop=self._loop,
            instance_name='foo',
            subscriptions_store=subscriptions_store,
            timeout=config.timeout,
            nosleep=nosleep)
        return self._reader


class When_a_stream_contains_a_single_event_and_the_counter_is_at_the_start(StreamReaderContext):

    _event = None

    @property
    def the_event(self):
        if(not self._event):
            self._event = self._queue.get_nowait()
        return self._event

    def given_a_feed_containing_one_event(self):
        self.http.registerJsonUri(
            'http://eventstore.local:2113/streams/newstream/0/forward/20',
            SCRIPT_PATH + '/responses/single-event.json')

    def because_we_start_the_reader(self):
        self.subscribe_and_run('newstream')

    def it_should_contain_the_body(self):
        assert(self.the_event.data == {"a": 1})

    def it_should_contain_the_event_type(self):
        assert(self.the_event.type == "my-event")

    def it_should_contain_the_event_id(self):
        assert(
            self.the_event.id == UUID('fbf4a1a1-b4a3-4dfe-a01f-ec52c34e16e4'))

    def it_should_contain_the_sequence_number(self):
        assert(self.the_event.sequence == 0)

    def it_should_contain_the_stream_id(self):
        assert(self.the_event.stream == "newstream")


class When_an_event_contains_no_data(StreamReaderContext):

    _event = None

    def given_a_feed_containing_one_event(self):
        self.http.registerJsonUri(
            'http://eventstore.local:2113/streams/newstream/0/forward/20',
            SCRIPT_PATH + '/responses/single-event-no-data.json')

    def because_we_start_the_reader(self):
        self._log = SpyLog()
        with(self._log.capture()):
            self.subscribe_and_run('newstream')

    def it_should_return_a_none_event(self):
        assert(self._queue.empty())

    def it_should_log_a_warning(self):
        for r in self._log._logs:
            print(r.msg)
        assert(any(r.msg.startswith("No `data` key found on event")
                   and r.levelno == logging.WARNING
                   for r in self._log._logs))

class When_a_feed_contains_multiple_events(StreamReaderContext):

    _the_events = []

    def given_a_feed_containing_three_events(self):
        self.http.registerJsonUri('http://eventstore.local:2113/streams/foo/0/forward/20',
                                  SCRIPT_PATH + '/responses/three-events.json')

    def because_we_start_the_reader(self):
        self.subscribe_and_run('foo')

    def it_should_return_three_events(self):
        assert(len(self.the_events) == 3)

    def the_events_should_be_ordered_correctly(self):
        assert(self.the_events[0].sequence == 0)
        assert(self.the_events[1].sequence == 1)
        assert(self.the_events[2].sequence == 2)

    @property
    def the_events(self):
        if(not self._the_events):
            while(not self._queue.empty()):
                self._the_events.append(self._queue.get_nowait())

        return self._the_events


class When_a_last_read_event_is_specified(StreamReaderContext):

    _the_events = []

    def given_a_feed_containing_three_events(self):
        self.http.registerJsonUri('http://eventstore.local:2113/streams/foo/1/forward/20',
                                  SCRIPT_PATH + '/responses/three-events.json')
        self.http.registerJsonUri(
            'http://127.0.0.1:2113/streams/newstream2/3/forward/20',
            SCRIPT_PATH + '/responses/two-page/head_next_prev_prev.json')

    def because_we_start_the_reader(self):
        self.subscribe_and_run('foo', last_read=1)

    def it_should_return_one_event(self):
        assert(len(self.the_events) == 1)

    def the_events_should_be_ordered_correctly(self):
        assert(self.the_events[0].sequence == 2)

    @property
    def the_events(self):
        if(not self._the_events):
            while(not self._queue.empty()):
                self._the_events.append(self._queue.get_nowait())

        return self._the_events


# For some reason (ie. because I deleted things), my stock feed got a bit messed up
# but that gives us a useful test case if nowt else
# class When_a_feed_spans_several_pages(StreamReaderContext):

#     _the_events = []

#     def given_a_feed_spanning_two_pages(self):
#         self._host = "127.0.0.1"

#         # all the pages except the first and last in this list are empty.
#         self.http.registerJsonUris(
#             {
#                 'http://127.0.0.1:2113/streams/stock':
#                     SCRIPT_PATH + '/responses/two-page/head.json',
#                 'http://127.0.0.1:2113/streams/stock/0/forward/20':
#                     SCRIPT_PATH + '/responses/two-page/head_last.json',
#                  'http://127.0.0.1:2113/streams/stock/20/forward/20':
#                     SCRIPT_PATH + '/responses/two-page/head_last_prev.json',
#                  'http://127.0.0.1:2113/streams/stock/40/forward/20':
#                     SCRIPT_PATH + '/responses/two-page/head_last_prev_prev.json',
#                  'http://127.0.0.1:2113/streams/stock/60/forward/20':
#                     SCRIPT_PATH + '/responses/two-page/head_last_prev_prev_prev.json',
#                  'http://127.0.0.1:2113/streams/stock/80/forward/20':
#                     SCRIPT_PATH + '/responses/two-page/head_last_prev_prev_prev_prev.json',
#                  'http://127.0.0.1:2113/streams/stock/84/forward/20':
#                     SCRIPT_PATH + '/responses/two-page/head_last_prev_prev_prev_prev_prev.json',
#             }
#         )

#     def because_we_start_the_reader(self):
#         self.subscribe_and_run('stock')

#     def we_should_raise_all_the_events(self):
#         assert(len(self.the_events) == 22)

#     def we_should_raise_the_events_in_the_correct_order(self):
#         assert(self.the_events[0].sequence == 62)
#         assert(self.the_events[21].sequence == 83)

#     @property
#     def the_events(self):
#         if(not self._the_events):
#             while(not self._queue.empty()):
#                 self._the_events.append(self._queue.get_nowait())

#         return self._the_events


class When_the_last_read_event_is_on_the_first_page(StreamReaderContext):

    _the_events = []

    def given_a_feed_spanning_two_pages(self):
        self._host = "127.0.0.1"

        self.http.registerJsonUris({
            'http://127.0.0.1:2113/streams/stock/80/forward/20':
            SCRIPT_PATH + '/responses/two-page/head.json',
            'http://127.0.0.1:2113/streams/stock/84/forward/20':
            SCRIPT_PATH + '/responses/two-page/head_next_prev_prev.json',


            })

    def because_we_start_the_reader(self):
        self.subscribe_and_run('stock', last_read=80)

    def we_should_raise_the_new_events(self):
        assert(len(self.the_events) == 3)

    def we_should_raise_the_events_in_the_correct_order(self):
        assert(self.the_events[0].sequence == 81)
        assert(self.the_events[2].sequence == 83)

    @property
    def the_events(self):
        if(not self._the_events):
            while(not self._queue.empty()):
                self._the_events.append(self._queue.get_nowait())

        return self._the_events


class When_the_reader_is_invoked_for_a_second_time(StreamReaderContext):

    _event = None

    @property
    def the_event(self):
        if(not self._event):
            self._event = self._queue.get_nowait()
        return self._event

    def given_a_reader_that_has_read_all_the_events(self):
        self.http.registerJsonUri(
            'http://eventstore.local:2113/streams/newstream/0/forward/20',
            SCRIPT_PATH + '/responses/single-event.json')
        self.http.registerJsonUri(
            'http://127.0.0.1:2113/streams/newstream/1/forward/20',
            SCRIPT_PATH + '/responses/empty.json')
        self.http.registerJsonUri(
            'http://eventstore.local:2113/streams/newstream/0/forward/20',
            SCRIPT_PATH + '/responses/single-event.json')
        self.http.registerJsonUri(
            'http://127.0.0.1:2113/streams/newstream/1/forward/20',
            SCRIPT_PATH + '/responses/empty.json')

        self._reader = self.subscribeTo('newstream', -1)
        self.run_the_reader()
        self._queue.get_nowait()

    def because_we_run_the_reader_a_second_time(self):
        self.run_the_reader()

    def it_should_not_return_any_events(self):
        assert(self._queue.empty())

    def run_the_reader(self):
        mock = self.http.getMock()
        with patch("aiohttp.request", new=mock):
            self._loop.run_until_complete(
                self._reader.start_consuming(once=True)
            )


"""
In this context, we create a new stream containing a single event (head.json)
When we poll the stream a second time there are no new events, so we walk to the
previous link(head_prev.json, then head_prev2.json)

When we poll the stream a third time, there are 23 new events. Since that's more
than a page worth, we should first seek the last_read event on the next page
(head_prev2_next) then walk backward until we find a new empty page
(head_prev2_next_prev_prev.json)

"""


class When_events_are_added_after_the_first_run(StreamReaderContext):

    _the_events = []

    def given_a_changing_feed(self):
        # When we first hit the stream, it returns a single event
        self.http.registerJsonUri('http://eventstore.local:2113/streams/stock/0/forward/20',
                                  SCRIPT_PATH + '/responses/new-events/head.json')

        self.http.registerJsonUri('http://eventstore.local:2113/streams/stock/0/forward/20',
                                  SCRIPT_PATH + '/responses/new-events/head.json')

        self.http.registerJsonUri('http://eventstore.local:2113/streams/stock/0/forward/20',
                                  SCRIPT_PATH + '/responses/new-events/head.json')

        # On the second invocation, we receive no new events
        # On the third invocation, we receive two pages of two events each
        self.http.registerJsonsUri(
            'http://127.0.0.1:2113/streams/stock/85/forward/20',
            [SCRIPT_PATH + '/responses/new-events/head_prev.json',
             SCRIPT_PATH + '/responses/new-events/head_prev2.json',
             SCRIPT_PATH + '/responses/new-events/head_prev2.json', ])
        self.http.registerJsonUri(
            'http://127.0.0.1:2113/streams/stock/84/backward/20',
            SCRIPT_PATH + '/responses/new-events/head_prev2_next.json')
        self.http.registerJsonUri(
            'http://127.0.0.1:2113/streams/stock/105/forward/20',
            SCRIPT_PATH + '/responses/new-events/head_prev2_next_prev.json')
        self.http.registerJsonUri(
            'http://127.0.0.1:2113/streams/stock/108/forward/20',
            SCRIPT_PATH + '/responses/new-events/head_prev2_next_prev_prev.json')

    def because_we_run_the_reader_three_times(self):
        self._reader = self.subscribeTo('stock', -1)

        self.run_the_reader()
        self.run_the_reader()
        self.run_the_reader()

    def it_should_have_read_all_the_events(self):
        assert(len(self.the_events) == 24)

    def it_should_have_read_the_events_in_the_correct_order(self):
        assert(self.the_events[0].sequence == 84)
        assert(self.the_events[23].sequence == 107)

    def run_the_reader(self):
        mock = self.http.getMock()
        with patch("aiohttp.request", new=mock):
            self._loop.run_until_complete(
                self._reader.start_consuming(once=True)
            )

    @property
    def the_events(self):
        if(not self._the_events):
            while(not self._queue.empty()):
                self._the_events.append(self._queue.get_nowait())

        return self._the_events



class When_reading_from_a_category_projection(StreamReaderContext):

    _the_events = []

    @property
    def the_events(self):
        if(not self._the_events):
            while(not self._queue.empty()):
                self._the_events.append(self._queue.get_nowait())

        return self._the_events

    def given_a_category_projection_stream(self):
        self.http.registerJsonUri(
            'http://eventstore.local:2113/streams/$ce-order/0/forward/20',
            SCRIPT_PATH + '/responses/category-projection/head.json'
        )

    def because_we_run_the_reader(self):
        self._reader = self.subscribeTo('$ce-order', -1)
        self.run_the_reader()

    def it_should_have_read_all_the_events(self):
        assert len(self.the_events) == 3

    def it_should_have_read_the_events_in_the_correct_order(self):
        for seq in range(3):
            assert self.the_events[seq].sequence == seq

    def it_should_use_the_category_projection_stream_name(self):
        for seq in range(3):
            assert self.the_events[seq].stream == '$ce-order'


"""
Value errors pretty much mean that our URL is screwed, or that there's an SSL
context mismatch. In that case, we should just end the loop.
"""


class When_a_valueerror_occurs_during_fetch(StreamReaderContext):

    _log = SpyLog()

    def given_a_malformed_port(self):
        self._port = "tawny"

    # note that we run with a real event loop, and don't explicitly
    # call stop. The exception will stop the loop.
    def because_we_start_the_reader(self):
        self._reader = self.subscribeTo("my-stream", 1)
        with(self._log.capture()):
            self._loop.run_until_complete(
                self._reader.start_consuming()
                )

    def it_should_log_a_critical_error(self):
        assert(filter(lambda r: r.level == logging.CRITICAL, self._log._logs))


"""
If we get a client error, then something has gone wrong with the http
layer processing. It's almost certainly an intermittent fault and we should
retry.
"""


class When_a_client_error_occurs_during_fetch(StreamReaderContext):

    _log = SpyLog()

    def given_a_client_error(self):
        self.http.registerCallbacksUri(
            'http://eventstore.local:2113/streams/newstream/0/forward/20',
            [
                lambda: exec('raise aiohttp.errors.ClientOSError("Darn it, can\'t connect")'),
                lambda: exec('raise ValueError()')
            ]
        )

    def because_we_start_the_reader(self):
        self._reader = self.subscribeTo("newstream", -1, nosleep=True)
        with(self._log.capture()):
            mock = self.http.getMock()
            with patch("aiohttp.request", new=mock):
                self._loop.run_until_complete(
                    self._reader.start_consuming()
                )

    def it_should_log_a_warning(self):
        for r in self._log._logs:
            print(r.msg)
        assert(any(r.msg == "Error occurred while requesting %s"
                   and r.levelno == logging.WARNING
                   for r in self._log._logs))


class When_multiple_errors_of_the_same_type_occur(StreamReaderContext):

    _log = SpyLog()

    def given_a_client_error(self):
        self.http.registerCallbacksUri(
            'http://eventstore.local:2113/streams/newstream/0/forward/20',
            [
                lambda: exec('raise aiohttp.errors.ClientOSError("Darn it, can\'t connect")'),
                lambda: exec('raise aiohttp.errors.ClientOSError("Darn it, can\'t connect")'),
                lambda: exec('raise ValueError()')
            ]
        )

    def because_we_start_the_reader(self):
        self._reader = self.subscribeTo("newstream", -1, nosleep=True)
        with(self._log.capture()):
            mock = self.http.getMock()
            with patch("aiohttp.request", new=mock):
                self._loop.run_until_complete(
                    self._reader.start_consuming()
                )

    def it_should_log_a_warning(self):
        for r in self._log._logs:
            print(r.msg)
        assert(len([r for r in self._log._logs if r.msg.startswith("Error occurred while requesting %s")]) == 1)


"""
If we get a disconnection error, then it's a network level issue. Retry with
a backoff.
"""


class When_a_disconnection_error_occurs_during_fetch(StreamReaderContext):

    _log = SpyLog()

    def given_a_disconnection_error(self):
        self.http.registerCallbacksUri(
            'http://eventstore.local:2113/streams/newstream/0/forward/20',
            [
                lambda: exec('raise aiohttp.errors.DisconnectedError("Darn it, can\'t connect")'),
                lambda: exec('raise ValueError()')
            ]
        )

    def because_we_start_the_reader(self):
        self._reader = self.subscribeTo("newstream", -1, nosleep=True)
        with(self._log.capture()):
            mock = self.http.getMock()
            with patch("aiohttp.request", new=mock):
                self._loop.run_until_complete(
                    self._reader.start_consuming()
                )

    def it_should_log_a_warning(self):
        assert(any(r.msg == "Error occurred while requesting %s"
                   and r.levelno == logging.WARNING
                   for r in self._log._logs))


class When_a_timeout_error_occurs_during_fetch(StreamReaderContext):

    """
    If we get a Timeout error, then it's a network level issue. Retry with
    a backoff.
    """

    _log = SpyLog()

    def given_a_timeout_error(self):
        self.http.registerCallbacksUri(
            'http://eventstore.local:2113/streams/newstream/0/forward/20',
            [
                lambda: exec('raise aiohttp.errors.TimeoutError()'),
                lambda: exec('raise ValueError()')
            ]
        )

    def because_we_start_the_reader(self):
        self._reader = self.subscribeTo("newstream", -1, nosleep=True)
        with(self._log.capture()):
            mock = self.http.getMock()
            with patch("aiohttp.request", new=mock):
                self._loop.run_until_complete(
                    self._reader.start_consuming()
                )

    def it_should_log_a_warning(self):
        assert(any(r.msg == "Error occurred while requesting %s"
                   and r.levelno == logging.WARNING
                   for r in self._log._logs))


"""
If we get a ClientResponseError error, then it's a network level issue. Retry with
a backoff.
"""


class When_a_client_response_error_occurs_during_fetch(StreamReaderContext):

    _log = SpyLog()

    def given_a_client_response_error(self):
        self.http.registerCallbacksUri(
            'http://eventstore.local:2113/streams/newstream/0/forward/20',
            [
                lambda: exec('raise aiohttp.errors.ClientResponseError("Darn it, something went bad")'),
                lambda: exec('raise ValueError()')
            ]
        )

    def because_we_start_the_reader(self):
        self._reader = self.subscribeTo("newstream", -1, nosleep=True)
        with(self._log.capture()):
            mock = self.http.getMock()
            with patch("aiohttp.request", new=mock):
                self._loop.run_until_complete(
                    self._reader.start_consuming()
                )

    def it_should_log_a_warning(self):
        assert(any(r.msg == "Error occurred while requesting %s"
                   and r.levelno == logging.WARNING
                   for r in self._log._logs))


"""
If we get a 40x range message then something is wrong with our configuration,
we should stop the loop and log an error.
"""


class When_we_receive_a_4xx_range_error(StreamReaderContext):

    _log = SpyLog()

    def given_a_415(self):
        self.http.registerEmptyUri(
            'http://eventstore.local:2113/streams/newstream/0/forward/20', 415)

    def because_we_fetch_the_stream(self):
        self._reader = self.subscribeTo("newstream", -1, nosleep=True)
        with(self._log.capture()):
            mock = self.http.getMock()
            with patch("aiohttp.request", new=mock):
                self._loop.run_until_complete(
                    self._reader.start_consuming()
                )

    def it_should_log_an_error(self):
        assert(
            any(r.msg == "Received bad http response with status %d from %s"
                and r.levelno == logging.ERROR
                for r in self._log._logs))


class When_we_receive_a_404_range_error(StreamReaderContext):

    _log = SpyLog()

    def given_a_404(self):
        self.http.registerEmptyUri(
            'http://eventstore.local:2113/streams/newstream/0/forward/20', 404)
        self.http.registerCallbackUri(
                'http://eventstore.local:2113/streams/newstream/0/forward/20',
                lambda: exec('raise ValueError()'))

    def because_we_fetch_the_stream(self):
        self._reader = self.subscribeTo("newstream", -1, nosleep=True)
        with(self._log.capture()):
            mock = self.http.getMock()
            with patch("aiohttp.request", new=mock):
                self._loop.run_until_complete(
                    self._reader.start_consuming()
                )

    def it_should_log_a_single_warning(self):
        assert(
            any(r.msg == "Error occurred while requesting %s"
                and r.levelno == logging.WARNING
                for r in self._log._logs))


class When_we_receive_a_408_range_error(StreamReaderContext):

    _log = SpyLog()

    def given_a_408(self):
        self.http.registerEmptyUri(
            'http://eventstore.local:2113/streams/newstream/0/forward/20', 408)
        self.http.registerCallbackUri(
                'http://eventstore.local:2113/streams/newstream/0/forward/20',
                lambda: exec('raise ValueError()'))

    def because_we_fetch_the_stream(self):
        self._reader = self.subscribeTo("newstream", -1, nosleep=True)
        with(self._log.capture()):
            mock = self.http.getMock()
            with patch("aiohttp.request", new=mock):
                self._loop.run_until_complete(
                    self._reader.start_consuming()
                )

    def it_should_log_a_single_warning(self):
        assert(
            any(r.msg == "Error occurred while requesting %s"
                and r.levelno == logging.WARNING
                for r in self._log._logs))


"""
50x range error means that something has gone wonky with event store. Log an error
and retry with a backoff.
"""


class When_we_receive_a_50x_range_error(StreamReaderContext):

    _log = SpyLog()

    def given_a_500(self):
        self.http.registerEmptyUri(
                'http://eventstore.local:2113/streams/newstream/0/forward/20', 500)
        self.http.registerCallbackUri(
                'http://eventstore.local:2113/streams/newstream/0/forward/20',
                lambda: exec('raise ValueError()'))

    def because_we_fetch_the_stream(self):
        self._reader = self.subscribeTo("newstream", -1, nosleep=True)
        with(self._log.capture()):
            mock = self.http.getMock()
            with patch("aiohttp.request", new=mock):
                self._loop.run_until_complete(
                    self._reader.start_consuming()
                )

    def it_should_log_a_single_warning(self):
        assert(
            any(r.msg == "Error occurred while requesting %s"
                and r.levelno == logging.WARNING
                for r in self._log._logs))


class When_an_event_has_bad_json(StreamReaderContext):

    _log = SpyLog()

    def given_a_feed_containing_one_event(self):
        self.http.registerJsonUri(
            'http://eventstore.local:2113/streams/newstream/0/forward/20',
            SCRIPT_PATH + '/responses/invalid-event.json')

    def because_we_start_the_reader(self):
        with(self._log.capture()):
            self.subscribe_and_run('newstream')

    def it_should_log_an_error(self):
        assert(any(
            r.msg == "Failed to parse json data for %s message %s"
            and r.levelno == logging.ERROR
            for r in self._log._logs))


class When_reading_from_a_stream_with_a_dynamic_date(StreamReaderContext):

    _event = None

    @property
    def the_event(self):
        if(not self._event):
            self._event = self._queue.get_nowait()
        return self._event

    def given_a_feed_for_today_containing_one_event(self):
        self.http.registerJsonUri(
            'http://eventstore.local:2113/streams/sotd_{}/0/forward/20'.format(
                datetime.date.today().isoformat()),
            '{}/responses/single-event.json'.format(SCRIPT_PATH))

    def because_we_start_the_reader(self):
        self.subscribe_and_run('sotd_#date#')

    def it_should_have_the_event_coming_from_the_stream_for_today(self):
        assert(
            self.the_event.id == UUID('fbf4a1a1-b4a3-4dfe-a01f-ec52c34e16e4'))


class When_reading_from_a_stream_with_a_dynamic_date_twice_across_dates(StreamReaderContext):

    def given_a_dynamic_subscription(self):
        self.http.registerJsonUri(
            'http://eventstore.local:2113/streams/sotd_2015-08-27/0/forward/20',
            '{}/responses/first_date_response.json'.format(SCRIPT_PATH))
        self.http.registerJsonUri(
            'http://eventstore.local:2113/streams/sotd_2015-08-28/0/forward/20',
            '{}/responses/second_date_response.json'.format(SCRIPT_PATH))

        self.subscribeTo("sotd_#date#", -1, nosleep=True)

    def because_we_start_the_reader_on_two_different_dates(self):
        with freeze_time("2015-08-27"):
            self.run_the_reader()

        with freeze_time("2015-08-28"):
            self.run_the_reader()

        self._first_event = self._queue.get_nowait()
        self._second_event = self._queue.get_nowait()

    def it_should_have_the_event_coming_from_the_stream_for_the_first_date(self):
        assert(
            self._first_event.id == UUID('a51be208-25e5-41b8-a598-ec299a20ef96'))

    def it_should_have_the_event_coming_from_the_stream_for_the_second_date(self):
        assert(
            self._second_event.id == UUID('fbf4a1a1-b4a3-4dfe-a01f-ec52c34e16e4'))



from setuptools import setup, find_packages

install_requires = [
    "aiohttp>=0.15.3",
    "asyncio==3.4.3",
    "chardet==2.3.0",
    "gevent>=1.1.1",
    "greenlet==0.4.5",
    "pybreaker==0.2.3",
    "PyYAML==3.11",
    "redis==2.10.3",
    "retrying==1.3.3",
    "requests==2.7.0"
]

tests_require = [
    "Contexts==0.10.2",
    "fakeredis==0.6.1",
    "freezegun==0.3.3",
    "HTTPretty"
]

setup(
    name="AtomicPuppy",
    version="0.2.4",
    packages=find_packages(),
    dependency_links=[
        "git+https://github.com/OddBloke/HTTPretty.git@f899d1bda8234658c2cec5aab027cb5b7c42203c#egg=HTTPretty"
    ],
    install_requires=install_requires,
    tests_require=tests_require,
    url='https://github.com/madedotcom/atomicpuppy',
    download_url='https://github.com/madedotcom/atomicpuppy/tarball/0.2.4',
    description='A service-activator component for eventstore',
    author='Bob Gregory',
    keywords=['eventstore'],
)



import asyncio
from aiohttp.client import ClientResponse
from unittest.mock import Mock
from collections import deque, defaultdict
import logging


class FakeHttp:

    def __init__(self, loop):
        self._loop = loop
        self._callbacks = defaultdict(deque)

    def registerJsonUris(self, registrations):
        for uri, fn in registrations.items():
            self.registerJsonUri(uri, fn)

    def _make_response(self, uri, filename):
        def cb():

            def read():
                fut = asyncio.Future(loop=self._loop)
                with open(filename) as f:
                    fut.set_result(f.read().encode('utf-8'))
                return fut

            resp = ClientResponse('GET', uri)
            resp.headers = {
                'Content-Type': 'application/json'
            }

            resp.status = 200
            resp.content = Mock()
            resp.content.read.side_effect = read
            resp.close = Mock()
            fut = asyncio.Future(loop=self._loop)
            fut.set_result(resp)
            return fut
        return cb

    def registerJsonUri(self, uri, filename):
        self._callbacks[uri].append(self._make_response(uri, filename))

    def registerJsonsUri(self, uri, filenames):
        for f in filenames:
            self._callbacks[uri].append(self._make_response(uri, f))

    def registerEmptyUri(self, uri, status):
        def cb():
            fut = asyncio.Future(loop=self._loop)
            resp = ClientResponse('GET', 'foo')
            resp.status = status
            fut.set_result(resp)
            return fut
        self._callbacks[uri].append(cb)

    def registerCallbackUri(self, uri, cb):
        self._callbacks[uri].append(cb)

    def registerCallbacksUri(self, uri, funcs):
        for cb in funcs:
            self._callbacks[uri].append(cb)

    @asyncio.coroutine
    def respond(self, uri):
        if(self._callbacks[uri]):
            cb = self._callbacks[uri].popleft()
            return cb()
        print("No response registered for uri "+uri)

    def getMock(self):
        return Mock(side_effect=lambda m, u, **kwargs: self.respond(u))


class SpyLog(logging.StreamHandler):

    def __init__(self):
        self._logs = []
        self.setLevel(logging.DEBUG)
        self.filters = []
        self.lock = None
        self.formatter = logging.Formatter()

    def emit(self, record):
        self.formatter.format(record)
        self._logs.append(record)

    def capture(self):
        return SpyLogContextManager(self)

    @property
    def errors(self):
        return [r for r in self._logs if r.levelname == "ERROR"]

    @property
    def warnings(self):
        return [r for r in self._logs if r.levelname == "WARNING"]


class SpyLogContextManager:

    def __init__(self, log):
        self._log = log

    def __enter__(self):
        logging.getLogger().addHandler(self._log)

    def __exit__(self, type, value, traceback):
        logging.getLogger().removeHandler(self._log)
        print("logs:")
        for r in self._log._logs:
            print(r.msg)



from uuid import uuid4

import httpretty

from atomicpuppy import Event, EventPublisher


class When_a_message_is_posted:

    stream = str(uuid4())
    event_id = str(uuid4())

    class TestConfig:

        def __init__(self, host, port):
            self.host = host
            self.port = port

    def given_a_publisher(self):
        cfg = When_a_message_is_posted.TestConfig('fakehost', '42')
        self.publisher = EventPublisher(cfg.host, cfg.port)

    @httpretty.activate
    def because_an_event_is_published_on_a_stream(self):
        httpretty.register_uri(
            httpretty.POST,
            "http://fakehost:42/streams/{}".format(self.stream),
            body='{}')

        data = {'foo': 'bar'}
        evt = Event(self.event_id, 'my-event-type', data, self.stream, None)
        self.publisher.post(evt)

    def it_should_be_a_POST(self):
        assert(httpretty.last_request().method == "POST")

    def it_should_have_sent_the_correct_id(self):
        assert(httpretty.last_request().headers['ES-EventId'] == self.event_id)

    def it_should_have_sent_the_correct_type(self):
        assert(httpretty.last_request().headers['ES-EventType'] == 'my-event-type')

    def it_should_have_sent_the_correct_body(self):
        assert(httpretty.last_request().body == b'{"foo": "bar"}')



class FatalError(Exception):
    pass


class UrlError(FatalError):

    def __init__(self, uri):
        self.uri = uri


class HttpClientError(FatalError):

    def __init__(self, uri, status):
        self.status = status
        self.uri = uri


class HttpServerError(Exception):

    def __init__(self, uri, status):
        self.status = status
        self.uri = uri


class RejectedMessageException(Exception):
    pass
